/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "libsnp.h"
#include "libsmm.h"
#include "snpusr.h"
#include "snp.h"
#include "emm.h"
//#include "gammaf.h"
//#include "igammaf.h"

using namespace liberr;
using namespace libcpp;
using namespace libsnp;
using namespace libsmm;
using namespace snp;
using namespace emm;
using namespace std;

namespace {
  const INTEGER n = 4750; //number of days in sample
  const INTEGER p = 2;   //numebr of parameters in beta
  const INTEGER nm = 2; //number of moments
  const INTEGER ML = 1; //maximum number of lags used
  const REAL trunc = 0.5124;
  const INTEGER indic_sign = -1; //indicates to look at positive or negative jumps; 1 is for positive and -1 for negative

  realmat  moment(const realmat& data, const realmat& jumps, const realmat& rho, INTEGER t, INTEGER NJ, const INTEGER& indic_sign)
   {
//NJ stands for number of jumps before this day
     //the parameters
     REAL k = rho[1];
     REAL sigma = rho[2]/100.0;
     INTEGER NJ_T = INTEGER(data(t,2)+data(t,3));
     realmat m(nm,1,0.0);

     REAL temp1 = 0.0; REAL temp2 = 0.0;
     for (INTEGER i=1; i<=NJ_T; i++) {
        if (indic_sign>0) { //look at POSITIVE JUMPS
          if (jumps[NJ+i]>0.0) {
            REAL jr = 100.0*(exp(jumps[NJ+i]/100.0)-exp(trunc/100.0));
            temp1 += -1.0/sigma+(1.0/k+1.0)*(k*jr/pow(sigma,2.0))/(1.0+k*jr/sigma);
            temp2 += log(1.0+k*jr/sigma)/pow(k,2.0)-(1.0/k+1.0)*(jr/sigma)/(1.0+k*jr/sigma);
          }
        }
        else { //look at NEGATIVE JUMPS
          if (jumps[NJ+i]<0.0) {
            REAL jr = 100.0*(exp(-jumps[NJ+i]/100.0)-exp(trunc/100.0));
            temp1 += -1.0/sigma+(1.0/k+1.0)*(k*jr/pow(sigma,2.0))/(1.0+k*jr/sigma);
            temp2 += log(1.0+k*jr/sigma)/pow(k,2.0)-(1.0/k+1.0)*(jr/sigma)/(1.0+k*jr/sigma);
          }
        }
     }

     m(1,1) = temp1; m(2,1) = temp2; 
//     m(3,1) = temp1*data(t-1,1); m(4,1) = temp2*data(t-1,1);

     return m;
   }

  REAL parzen(REAL x)
    {
      REAL z = fabs(x);
      if (z <= 0.5) {
        return 1.0 - 6.0*pow(z,2) + 6.0*pow(z,3);
       }
      if (z <= 1.0) {
        return 2.0*pow((1.0 - z),3);
       }
    return 0.0;
  }


};

emm::iv_usrmod::iv_usrmod
  (const std::vector<std::string>& usrprob, std::ostream& detail)
{
  beta.resize(p,1,0.0);
}

emm::iv_objfun::iv_objfun(const char* filename, std::ostream& detail)
{
  detail << starbox("/The model//") << '\n';  
  detail.flush();
  realmat dat;
  vecread("iv_data_3.dat",dat); //data contains the observations- the COLUMNS is the number of observations
  data.resize(n,3,0.0);
  for (INTEGER i=1; i<=n; i++) {
    data(i,1)=dat(1,i); data(i,2)=dat(2,i); data(i,3)=dat(3,i);
  }
  realmat temp_jumps;
  vecread("jm_data_3.dat",temp_jumps);
  jumps = temp_jumps;
  beta.resize(p,1,0.0);  
  W.resize(nm,nm,0.0);
  for (INTEGER i=1; i<=nm; i++){
    W(i,i)=1.0;
  }

  detail << starbox("/Parameter settings//") << '\n';
  detail << "\t n = " << n << '\n';
  detail << "\t p = " << p << '\n';
  //detail << "\t beta = " << beta << '\n';
  detail.flush();

  detail << starbox("/First six observations on data and the weighting matrix//") << '\n';
  detail << jumps("1:5"," ") << '\n';
  detail << W << '\n';
  detail.flush();
}

emm::iv_objfun::iv_objfun() 
: data(), jumps(), W(), beta()
{ error("Error, iv_objfun, default constructor got called"); }

emm::iv_objfun::iv_objfun(const iv_objfun& obj)
: data(obj.data), jumps(obj.jumps), W(obj.W), beta(obj.beta)
{ }

iv_objfun& emm::iv_objfun::operator=(const iv_objfun& obj)
{
  if (this != &obj) {
    data = obj.data; W = obj.W;  beta = obj.beta; jumps = obj.jumps;
  }
  return *this;
}

void emm::iv_objfun::set_parms(const realmat& b)
{
  if (b.get_rows() == beta.get_rows() && b.get_cols() == beta.get_cols()) {
    beta = b;
  }
  else {
    error("Error, iv_objfun, set_parms, wrong diminsions");
  }
}

REAL emm::iv_objfun::operator()(const realmat& b) const
{
  INTEGER pos_neg = 2; if (indic_sign<0) {pos_neg = 3;} //this is todetermine which jumps to count
  realmat moments(n,nm,0.0);
  INTEGER NJ = 0; INTEGER NSJ = 0; //NJ is TOTAl number of jumps and NSJ is number of SIGNED jumps
  for (INTEGER t=1; t<=ML; t++) {
     NJ = NJ+INTEGER(data(t,2)+data(t,3));  
     NSJ = NSJ + INTEGER(data(t,pos_neg));
  }
  for (INTEGER t=ML+1; t<=n; t++) {
    realmat temp = moment(data,jumps,b,t,NJ,indic_sign);
    NJ = NJ+INTEGER(data(t,2)+data(t,3));
    NSJ = NSJ+INTEGER(data(t,pos_neg));
    for (INTEGER i=1; i<=nm; i++) {
      moments(t,i) += temp[i];
    }
  } 
  realmat mom_all(nm,1,0.0);
  for (INTEGER i=1; i<=nm; i++) {
     for (INTEGER t=ML+1; t<=n; t++) {
        mom_all[i] += moments(t,i)/REAL(NSJ);
     }
  }


  realmat R0(nm,nm,0.0);
  for (INTEGER t=ML+1; t<=n; t++){
    for (INTEGER i=1; i<=nm; i++) {
       for (INTEGER j=1; j<=nm; j++) {
         if (abs(moments(t,i))>0) {//only if there was a jump otherwise we will be substracting the mean in the days without jump! 
           R0(i,j) += (moments(t,i)-mom_all[i])*(moments(t,j)-mom_all[j]);
         }
       }
    }
  }
  R0 = R0/(NSJ);
  realmat W_1 = R0; W_1 = invpsd(W_1);

  //the objective function
  realmat obj = T(mom_all)*W_1*mom_all;
 
  REAL v = 0.5*REAL(NSJ)*obj[1];
  return v;
}  

objfun_base* emm::iv_objfun::new_objfun() 
{ 
  objfun_base* r = new(nothrow) iv_objfun(*this); 
  if (r == 0) error("Error, iv_objfun, operator new failed");
  return r; 
}

void emm::iv_objfun::delete_objfun(objfun_base* objfun_ptr)
{ 
  delete objfun_ptr; 
}

