#ifndef __FILE_EMMUSR_H_SEEN__

/*-----------------------------------------------------------------------------

Copyright (C) 2005.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "libsmm.h"

namespace emm {

  class iv_objfun;
  class iv_usrmod;

  typedef iv_objfun usr_objfun;
  typedef iv_usrmod usrmod_type;
    
  class iv_objfun : public libsmm::objfun_base {
  private:
    realmat data;
    realmat jumps;
    realmat W;
    realmat beta;
  public:
                         iv_objfun(const char* filename, std::ostream& detail);
                         iv_objfun();
                         iv_objfun(const iv_objfun& obj);
    iv_objfun&           operator=(const iv_objfun& obj);
    void                 set_parms(const realmat& beta); 
    REAL                 operator()(const realmat& beta) const;
    libsmm::objfun_base* new_objfun();
    void                 delete_objfun(libsmm::objfun_base* objfun_ptr);
  };

  class iv_usrmod : public libsmm::usrmod_base {
  private:
    realmat beta;
  public:
    iv_usrmod(const std::vector<std::string>& usrprob, std::ostream& detail);
    INTEGER len_rho() {return beta.size();}
    INTEGER len_stats() {return beta.size();}
    bool gen_sim(realmat& sim) {sim=beta; return true;}
    bool gen_sim(realmat& sim, realmat& stats) {sim=stats=beta; return true;}
    bool gen_bootstrap(realmat& sim) {sim=beta; return true;}
    void get_rho(realmat& rho) {rho = beta;}
    void set_rho(const realmat& rho) {beta = rho;}
    bool support(const realmat& rho){
    if (rho[1] <  0.0) return false;
    if (rho[2] <  0.0) return false;
    return true;
    }
    libsmm::den_val prior(const realmat& rho, const realmat& stats)
      {return libsmm::den_val(true,0.0);}
    void write_usrvar(const char* filename) {return;}
  };  

}

#define __FILE_EMMUSR_H_SEEN__
#endif
