#ifndef __FILE_SNPUSR_H_SEEN__

/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "realmat.h"
#include "snp_base.h"
#include "libsnp.h"

namespace snp {
  
  class datread;                      // These are the classes defined here.
  class resmusig;
  class plot;
  class simulate;
  class leverage;
  
  typedef datread   datread_type;     // The type names are used in snp.cpp.
  typedef resmusig  residual_type;    // Change the class reference to 
  typedef resmusig  mean_type;        // substitute user written classes. 
  typedef resmusig  variance_type;    // Leave the distribution code intact.
  typedef plot      plot_type;        // Hopefully the constructor call in
  typedef simulate  simulate_type;    // snp.cpp is general enough that 
  typedef leverage  user_type;        // snp.cpp changes will be unnecessary.
    
  class datread : public datread_base {
  private:
    datparms dpm;
  public:
    datread();
    void initialize(datparms dp);
    bool read_data(realmat& data);
  };
  
  class resmusig : public ancillary_base {
  private:
    optparms opm;
    datparms dpm;
    tranparms tpm;
    libsnp::snpden f;
    libsnp::afunc af;
    libsnp::ufunc uf;
    libsnp::rfunc rf;
    std::ostream& detail;
    const trnfrm* tr;
    const realmat* X;
    const realmat* Y;
    std::ostream* os;
  public:
    resmusig(optparms op, datparms dp, tranparms tp,
         libsnp::snpden fn, libsnp::afunc afn,
         libsnp::ufunc ufn, libsnp::rfunc rfn,
         std::ostream& dos, const trnfrm* trn);
    void set_XY(const realmat* x, const realmat* y);
    bool initialize(std::ostream* out_stream);
    bool calculate();
    bool finalize();
  };
  
  class simulate : public ancillary_base {
  private:
    optparms opm;
    datparms dpm;
    tranparms tpm;
    libsnp::snpden f;
    libsnp::afunc af;
    libsnp::ufunc uf;
    libsnp::rfunc rf;
    std::ostream& detail;
    const trnfrm* tr;
    const realmat* X;
    const realmat* Y;
    std::ostream* os;
  public:
    simulate(optparms op, datparms dp, tranparms tp,
         libsnp::snpden fn, libsnp::afunc afn,
         libsnp::ufunc ufn, libsnp::rfunc rfn,
         std::ostream& dos, const trnfrm* trn);
    void set_XY(const realmat* x, const realmat* y);
    bool initialize(std::ostream* out_stream);
    bool calculate();
    bool finalize();
  };
  
  class plot : public ancillary_base {
  private:
    optparms opm;
    datparms dpm;
    tranparms tpm;
    libsnp::snpden f;
    libsnp::afunc af;
    libsnp::ufunc uf;
    libsnp::rfunc rf;
    std::ostream& detail;
    const trnfrm* tr;
    const realmat* X;
    const realmat* Y;
    std::ostream* os;
  public:
    plot(optparms op, datparms dp, tranparms tp,
         libsnp::snpden fn, libsnp::afunc afn,
         libsnp::ufunc ufn, libsnp::rfunc rfn,
         std::ostream& dos, const trnfrm* trn);
    void set_XY(const realmat* x, const realmat* y);
    bool initialize(std::ostream* out_stream);
    bool calculate();
    bool finalize();
  };
  
  class leverage : public ancillary_base {
  private:
    optparms opm;
    datparms dpm;
    tranparms tpm;
    libsnp::snpden f;
    libsnp::afunc af;
    libsnp::ufunc uf;
    libsnp::rfunc rf;
    std::ostream& detail;
    const trnfrm* tr;
    const realmat* X;
    const realmat* Y;
    std::ostream* os;
  public:
    leverage(optparms op, datparms dp, tranparms tp,
         libsnp::snpden fn, libsnp::afunc afn,
         libsnp::ufunc ufn, libsnp::rfunc rfn,
         std::ostream& dos, const trnfrm* trn);
    void set_XY(const realmat* x, const realmat* y);
    bool initialize(std::ostream* out_stream);
    bool calculate();
    bool finalize();
  };
  
}

#define __FILE_SNPUSR_H_SEEN__
#endif
