#include <sstream>
#include <cerrno>
#include "libcpp.h"

using namespace libcpp;
using namespace std;


realmat vol(REAL& V_1, REAL& V_2, realmat& params, INT_32BIT& seed, INTEGER& freq, INTEGER& neul)
 {
   REAL theta_1 = params[1]; REAL kappa_1 = params[2]; REAL sigma_1 = params[3];
   REAL theta_2 = params[4]; REAL kappa_2 = params[5]; REAL sigma_2 = params[6];

   realmat output(freq,4,0.0); REAL ninc = REAL(freq*neul); REAL ninc_root = pow(REAL(freq*neul),0.5);
   for (INTEGER t=1; t<=freq; t++) {
     REAL iv = 0.0; REAL ret = 0.0;
     for (INTEGER i=1; i<=neul; i++) {
       ret += pow(V_1+V_2,0.5)*unsk(seed)/ninc_root;
       iv += (V_1+V_2)/ninc;
       V_1 += kappa_1*(theta_1-V_1)/ninc+sigma_1*V_1*unsk(seed)/ninc_root;
       V_2 += kappa_2*(theta_2-V_2)/ninc+sigma_2*V_2*unsk(seed)/ninc_root;
     }
     output(t,1) = iv; output(t,2) = ret;
   }
   output(1,3)=V_1; output(1,4)=V_2; //end value of spot vol - which is beginning fro next day
   return output;
 }

INTEGER loc(REAL& k_0, REAL& k_1, realmat& vol, REAL& obs)
 {
   INTEGER T = vol.get_rows(); REAL iv = 0.0;
   INTEGER t = 0; REAL intens = 0.0;
   while (obs>intens && t<=T-1) {
    intens += k_0/REAL(T)+k_1*iv;
    t++;
    iv=vol(t,1);
   }
   if(t==0) {t=1;}
   return t; //the returned value will be from 1 to T
 }


realmat levy(realmat& params, REAL& k_0, REAL& k_1, INT_32BIT& seed, realmat& vol)
 {
   REAL c_0 = params[1]; REAL b_0 = params[2]; REAL c_1 = params[3]; REAL b_1 = params[4]; REAL thre = params[5];
   INTEGER T = vol.get_rows(); REAL TERM = 0.0; //the business time (i.e. the time-varying intensity)
   realmat output(T,1,0.0);

   for (INTEGER t=1; t<=T; t++) {
     TERM += vol(t,1);
   }
   TERM = k_0+TERM*k_1; //this is simulation for one day

   REAL z = 100.0*(exp(thre/100.0)-1.0); REAL tol = 2.0*c_0*pow(z,-b_0)/b_0;  REAL gam= -log(ran(seed));
   while(gam<tol*TERM){
       gam -=log(ran(seed));
       REAL shot = z*pow(1.0-ran(seed),-1.0/b_0);
       shot = 100.0*log(shot/100.0+1.0);
       REAL indic=unsk(seed);  if(indic<0.0) {shot*=-1.0;}
       REAL loc_draw = ran(seed)*TERM; INTEGER local = loc(k_0, k_1, vol, loc_draw); output[local]+=shot;
   }

   tol = 2.0*c_1*pow(z,-b_1)/b_1;  gam= -log(ran(seed));
   while(gam<tol*TERM){
       gam -=log(ran(seed));
       REAL shot = z*pow(1.0-ran(seed),-1.0/b_1);
       shot = 100.0*log(shot/100.0+1.0);
       REAL indic=unsk(seed);  if(indic<0.0) {shot*=-1.0;}
       REAL loc_draw = ran(seed)*TERM; INTEGER local = loc(k_0, k_1, vol, loc_draw); output[local]+=shot;
   }
   return output;
 } 

realmat ret(realmat& par_vol, realmat& par_lev, REAL& k_0, REAL& k_1, INT_32BIT& seed_vol, INT_32BIT& seed_lev, INTEGER& T, INTEGER& freq, INTEGER& neul)
 {
   realmat output(T*freq,1,0.0); REAL V_1=par_vol[1]; REAL V_2=par_vol[4]; //start at mean value
   realmat temp(T,3,0.0);
   for (INTEGER t=1; t<=T; t++) {
     realmat cont_incr = vol(V_1, V_2, par_vol, seed_vol, freq, neul);
     realmat jump_incr = levy(par_lev, k_0, k_1, seed_lev, cont_incr);
     for (INTEGER i=1; i<=freq; i++) {
     if (jump_incr(i,1)<-0.3) {temp(t,1) =cont_incr(i,2); temp(t,2) =jump_incr(i,1); temp(t,3) = cont_incr(i,1);} 
       output[(t-1)*freq+i] = cont_incr(i,2)+jump_incr(i,1);
     }
     V_1 = cont_incr(1,3); V_2 = cont_incr(1,4);
   }
   vecwrite("temp.dat",temp);
   return output;
 }
