/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "libcpp.h"
#include "libsnp.h"
#include "libsmm.h"
#include "snp.h"
#include "emm.h"

using namespace liberr;
using namespace libcpp;
using namespace libsnp;
using namespace libsmm;
using namespace snp;
using namespace emm;
using namespace std;

namespace {

  void output(const estparms& epm,ostream& detail,INTEGER ifile,string prefix,
      const realmat& rho_sim, const realmat& stats_sim, const realmat& pi_sim,
      REAL reject, const realmat& rho_hat, const realmat& V_hat, INTEGER n,
      const realmat& rho_mean, const realmat& rho_mode, REAL post_high,
      const realmat& foc_hat, const realmat& I, const realmat& invJ, 
      INTEGER reps)
  {
    if (ifile > 999) error("Error, emm, output, nfile too big");

    string filename;
  
    /*
    filename = prefix + ".rho_hat.dat";
    vecwrite(filename.c_str(), rho_hat);
    */

    filename = prefix + ".rho_mean.dat";
    vecwrite(filename.c_str(), rho_mean);
  
    filename = prefix + ".rho_mode.dat";
    vecwrite(filename.c_str(), rho_mode);
  
    realmat V_hat_hess = invJ/n;

    filename = prefix + ".V_hat_hess.dat";
    vecwrite(filename.c_str(), V_hat_hess);
  
    realmat V_hat_info;

    if (!epm.kilse) {
      
      filename = prefix + ".V_hat_sand.dat";
      vecwrite(filename.c_str(), V_hat);

      V_hat_info = inv(I)/n;

      filename = prefix + ".V_hat_info.dat";
      vecwrite(filename.c_str(), V_hat_info);

      if (epm.print) {
        detail << starbox("/Get symptotics/results are cumulative//") << '\n';
        detail << "\t ifile = " << ifile << '\n';
        detail << '\n';
        detail << "\t rho_hat = " << rho_hat << '\n';
        detail << "\t V_hat = " << V_hat << '\n';
        detail << "\t n = " << n << '\n';
        detail << "\t rho_mean = " << rho_mean << '\n';
        detail << "\t rho_mode = " << rho_mode << '\n';
        detail << "\t post_high = " << post_high << '\n';
        detail << "\t I = " << I << '\n';
        detail << "\t invJ = " << invJ << '\n';
        detail << "\t foc_hat = " << foc_hat << '\n';
        detail << "\t reps = " << reps << '\n';
        detail.flush();
      }
  
    }
  
    filename = prefix + ".summary.dat";
    ofstream summary_ofs(filename.c_str());
    if (summary_ofs) {
      summary_ofs 
	/*
        << "      rhohat"
	*/
        << "     rhomean"
        << "     rhomode"
        << "      sesand"
        << "      sehess"
        << "      seinfo"
        << '\n';
      for (INTEGER i=1; i<=rho_hat.size(); ++i) {
        if (epm.kilse) {
          summary_ofs
	    /*
            << fmt('g',12,5,rho_hat[i])
	    */
            << fmt('g',12,5,rho_mean[i])
            << fmt('g',12,5,rho_mode[i])
            << "            "
            << fmt('g',12,5,sqrt(V_hat_hess(i,i)))
            << "            "
            << '\n';
        }
        else {
          summary_ofs
	    /*
            << fmt('g',12,5,rho_hat[i])
	    */
            << fmt('g',12,5,rho_mean[i])
            << fmt('g',12,5,rho_mode[i])
            << fmt('g',12,5,sqrt(V_hat(i,i)))
            << fmt('g',12,5,sqrt(V_hat_hess(i,i)))
            << fmt('g',12,5,sqrt(V_hat_info(i,i)))
            << '\n';
        }
      }
      summary_ofs << '\n';
      summary_ofs << "The log posterior (log prior - objfun) at the mode ";
      summary_ofs << "is" << fmt('g',12,5,post_high) << ".\n\n";
      summary_ofs << "For EMM where objfun.prior returns only 0 or 1 "; 
      summary_ofs << "this is a chi-square \non ltheta - 1 - lrho ";
      summary_ofs << "degrees freedom.  See SNP parmfile for ltheta.\n";
    }

    char number[5];
    sprintf(number,"%03d",ifile);

    filename = prefix + ".reject" + number + ".dat"; 
    ofstream reject_ofs(filename.c_str());
    if (reject_ofs) reject_ofs << reject << '\n';
    
    filename = prefix + ".rho" + number + ".dat"; 
    vecwrite(filename.c_str(), rho_sim);
    
    filename = prefix + ".stats" + number + ".dat"; 
    vecwrite(filename.c_str(), stats_sim);
    
    filename = prefix + ".pi" + number + ".dat"; 
    vecwrite(filename.c_str(), pi_sim);
  }
}

int main(int argc, char** argp, char** envp)
{

  istream* ctrl_ptr;
  if (argc == 2) {
    ctrl_ptr = new(nothrow) ifstream(argp[1]);
    string msg = "Error, emm, " + string(argp[1]) + " open failed";
    if( (ctrl_ptr == 0) || (!*ctrl_ptr) ) error(msg);
  }
  else {
    ctrl_ptr = new(nothrow) ifstream("control.dat");
    string msg = "Error, emm, control.dat open failed";
    if( (ctrl_ptr == 0) || (!*ctrl_ptr) ) error(msg);
  }
  istream& ctrl = *ctrl_ptr;

  string parmfile;
  string prefix;
 
  while(ctrl >> parmfile >> prefix) {

    string filename = prefix + ".detail.dat";

    ofstream detail_ofs(filename.c_str());
    if(!detail_ofs) error("Error, emm, detail.dat open failed");
    ostream& detail = detail_ofs;

    emmparms pf;

    if(!pf.read_parms(parmfile.c_str(),detail)) {
       detail.flush();
       error("Error, emm, cannot read EMM parmfile");
    }

    usrmod_type usrmod(pf.get_usrparms(), detail);

    proposal_base* proposal_ptr;

    if (pf.get_estparms().proptype == 0) {
      proposal_ptr = new(nothrow) group_move(pf.get_prop_groups());
      if (proposal_ptr == 0) error("Error, emm main, operator new failed.");
    }
    else if (pf.get_estparms().proptype == 1) {
      bool print = pf.get_estparms().print;
      proposal_ptr 
        = new(nothrow) conditional_move(pf.get_prop_groups(),detail,print);
      if (proposal_ptr == 0) error("Error, emm main, operator new failed.");
    } else {
      error("Error, emm, no user proposal type implemented");
      proposal_ptr = new(nothrow) group_move(pf.get_prop_groups());
      if (proposal_ptr == 0) error("Error, emm main, operator new failed.");
    }

    proposal_base& proposal = *proposal_ptr;

    objfun_base* objfun_ptr;

    if (pf.get_estparms().objtype == 0) {

      if (pf.get_estparms().print) {
        detail << starbox("/SNP now controls print//");
        detail.flush();
      }

      objfun_ptr 
        = new(nothrow) emm_objfun(pf.get_estparms().score.c_str(), detail);
      if (objfun_ptr == 0) error("Error, emm main, operator new failed.");

      if (pf.get_estparms().print) {
        detail << starbox("/EMM now controls print//");
        detail.flush();
      }
    }
    else if (pf.get_estparms().objtype == 1) {

      if (pf.get_estparms().print) {
        detail << starbox("/SNP now controls print//");
        detail.flush();
      }

      objfun_ptr 
        = new(nothrow) edf_objfun(pf.get_estparms().score.c_str(), detail);
      if (objfun_ptr == 0) error("Error, emm main, operator new failed.");

      if (pf.get_estparms().print) {
        detail << starbox("/EMM now controls print//");
        detail.flush();
      }
    }
    else {
      objfun_ptr 
        = new(nothrow) usr_objfun(pf.get_estparms().score.c_str(), detail);
      if (objfun_ptr == 0) error("Error, emm main, operator new failed.");
    }

    objfun_base& objfun = *objfun_ptr;

    realmat increment = pf.get_rho_increment();

    grid_likehood likehood(usrmod, objfun, increment);

    filename = prefix + ".emmcache.dat";
    if (!likehood.read_cache(filename.c_str())) { 
      warn("Warning, emm, emmcache read failed, using default instead");
    }

    smm_mcmc mcmc(proposal, likehood, usrmod);
    mcmc.set_simulation_size(pf.get_estparms().lsim);
    mcmc.set_temp(pf.get_estparms().temperature);

    realmat data = pf.get_data();
  
    asymptotics_base* asymptotics_ptr;

    if (pf.get_estparms().kilse) {
      asymptotics_ptr 
        = new(nothrow) minimal_asymptotics(data,usrmod,mcmc);
      if (asymptotics_ptr == 0) error("Error, emm main, operator new failed.");
    }
    else {
      asymptotics_ptr 
        = new(nothrow) bstrap_asymptotics(data, usrmod, objfun, mcmc);
      if (asymptotics_ptr == 0) error("Error, emm main, operator new failed.");
    }

    asymptotics_base& asymptotics = *asymptotics_ptr;

    INT_32BIT seed = pf.get_estparms().seed;
    realmat rho = pf.get_rho();

    realmat rho_sim;
    realmat stats_sim;
    realmat pi_sim;
    REAL reject;

    for (INTEGER ifile = 0; ifile <= pf.get_estparms().nfile; ++ifile) {

      reject = mcmc.draw(seed, rho, rho_sim, stats_sim, pi_sim);

      filename = prefix + ".emmcache.new";
      if (!likehood.write_cache(filename.c_str())) {
        warn("Warning, emm, emmcache write failed");
      }

      realmat rho_hat;
      realmat V_hat;
      INTEGER n;
  
      realmat rho_mean;
      realmat rho_mode;
      REAL post_high;
      realmat foc_hat;
      realmat I;
      realmat invJ;
      INTEGER reps = 0;
  
      asymptotics.set_asymptotics(rho_sim);
      asymptotics.get_asymptotics(rho_hat,V_hat,n);
      asymptotics.get_asymptotics(rho_mean,rho_mode,post_high, 
                                    I,invJ,foc_hat,reps);
    
      usrmod.set_rho(rho_mode);
      realmat usr_sim;
      usrmod.gen_sim(usr_sim);
      objfun(usr_sim);
      filename = prefix + ".diagnostics.dat";
      objfun.write_diagnostics(filename.c_str());
      
      likehood.get_grid(rho_mode);      // So discretization is apparent
      pf.write_parms(parmfile, prefix, rho_mode, invJ/n);
      usrmod.set_rho(rho_mode);
      filename = prefix + ".usrvar.dat";
      usrmod.write_usrvar(filename.c_str());
  
      likehood.get_grid(rho_sim);       // So discretization is apparent
        
      output(pf.get_estparms(), detail, ifile, prefix,
        rho_sim, stats_sim, pi_sim, reject, rho_hat, V_hat, n, 
        rho_mean, rho_mode, post_high, foc_hat, I, invJ, reps);
    }
    delete objfun_ptr;
    delete proposal_ptr;
    delete asymptotics_ptr;
  }
  delete ctrl_ptr;
  return 0;
}

