#ifndef __FILE_EMM_H_SEEN__

/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "libsnp.h"
#include "libsmm.h"
#include "emm_base.h"
#include "emmusr.h"
#include "snp.h"

namespace emm {
    
  class emmparms {
  private:
    std::vector<std::string> history;
    estparms epm;
    realmat data;
    realmat rho;
    intvec rho_mask;
    realmat rho_increment;
    realmat prop_scale;
    libsmm::prop_def prop_groups; 
    std::vector<std::string> usrparms;
    std::vector<std::string> grouptxt;
    std::vector<std::string> parmrhs;
    std::vector<std::string> scalerhs;
    std::vector<std::string> incrhs;
    bool rv;
  public:
    emmparms();
    bool read_parms(const char* filename, std::ostream& detail);
    bool write_parms(std::string parmfile, std::string prefix, 
           const realmat& mu, const realmat& sig) const;
    void set_estparms(const estparms& e);
    void set_rho(const realmat& r);
    const estparms& get_estparms() const {return epm;}
    const realmat& get_data() const {return data;}
    const realmat& get_rho() const {return rho;}
    const intvec& get_rho_mask() const {return rho_mask;}
    const realmat& get_rho_increment() const {return rho_increment;}
    const realmat& get_prop_scale() const {return prop_scale;}
    const libsmm::prop_def& get_prop_groups() const {return prop_groups;}
    const std::vector<std::string>& get_usrparms() const {return usrparms;}
  };
  
  class emm_objfun : public libsmm::objfun_base {
  private:
    snp::trnfrm tr;
    snp::snpll ll;
    realmat V;
    INTEGER iter;
    REAL    toler;
    realmat mutable mean_score;
    realmat rho_infmat;
  public:
                         emm_objfun(const char* filename, std::ostream& detail);
                         emm_objfun();
                         emm_objfun(const emm_objfun& obj);
                         ~emm_objfun() { }
    emm_objfun&          operator=(const emm_objfun& obj);
    void                 set_parms(const realmat& data); 
    void                 set_iter(INTEGER itr) {iter = itr;}
    void                 set_toler(REAL tol) {toler = tol;}
    REAL                 operator()(const realmat& sim) const;
    bool                 write_diagnostics(const char* filename) const;
    libsmm::objfun_base* new_objfun();
    void                 delete_objfun(libsmm::objfun_base* objfun_ptr);
  };

  class edf_objfun : public libsmm::objfun_base {
  private:
    realmat data;
    realmat edf_dat;
    INTEGER d;
    INTEGER lag;
    realmat mutable edf_sim;
  public:
                         edf_objfun(const char* filename, std::ostream& detail);
                         edf_objfun();
                         edf_objfun(const edf_objfun& obj);
                         ~edf_objfun() { }
    edf_objfun&          operator=(const edf_objfun& obj);
    void                 set_parms(const realmat& data); 
    REAL                 operator()(const realmat& sim) const;
    bool                 write_diagnostics(const char* filename) const;
    libsmm::objfun_base* new_objfun();
    void                 delete_objfun(libsmm::objfun_base* objfun_ptr);
  };

}
#define __FILE_EMM_H_SEEN__
#endif
