#ifndef __FILE_SNP_BASE_H_SEEN__

/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "realmat.h"

namespace snp {
  
  struct optparms {
    std::string pname;
    REAL snpver;
    INTEGER itmax0;
    INTEGER itmax1;
    REAL toler;
    bool print;
    INTEGER task;
    INTEGER extra;
    REAL sfac;
    INT_32BIT iseed;
    INTEGER ngrid;
    bool kilse;
  };
  
  struct datparms {
    INTEGER M;
    INTEGER n;
    INTEGER drop;
    INTEGER cond;
    bool reread;
    std::string dsn;
    intvec fields;
  };
  
  struct tranparms {
    bool useold;
    bool diag;
    INTEGER squash;
    REAL inflec;
    realmat mean;
    realmat variance;
  };
  
  class trnfrm {
  private:
    datparms dpm;
    tranparms tpm;
    realmat R;
    realmat P;
    REAL detP;
  public:
    trnfrm(const datparms& dp, const tranparms& tp, const realmat& data);
    trnfrm();
    ~trnfrm() { }
    trnfrm(const trnfrm& tr);
    trnfrm& operator=(const trnfrm& tr);
    const datparms& get_datparms() const {return dpm;}
    const tranparms& get_tranparms() const {return tpm;}
    const realmat& get_R() const {return R;}
    const realmat& get_P() const {return P;}
    REAL get_detP() const {return detP;}
    void set_n(INTEGER n) {dpm.n=n;}
    void normalize(realmat& y) const;
    void unnormalize(realmat& y) const;
    void unscale(realmat& sigma) const;
    void spline(realmat& x) const;
    void logistic(realmat& x) const;
  };
  
  class datread_base {
  public:
    virtual void initialize(datparms dpm) = 0;
    virtual bool read_data(realmat& data) = 0;
    virtual ~datread_base() { }
  };
  
  class ancillary_base {
  public:
    virtual void set_XY(const realmat* x, const realmat* y) = 0;
    virtual bool initialize(std::ostream* out_stream) = 0;
    virtual bool calculate() = 0;
    virtual bool finalize() = 0;
    virtual ~ancillary_base() { }
  };
  
}

#define __FILE_SNP_BASE_H_SEEN__
#endif
